#include "interpreter.hh"
#include "customer.hh"
#include "command.hh"
#include "mall.hh"
#include "world.hh"
#include "parser.hh" // for LOAD
#include <iostream>
#include <map>
#include <stdexcept>
#include <fstream>

using namespace std;

class UserCommand: public Command {
public:
  UserCommand(Customer& customer)
    : _unexecutable(true), _customer(customer) {}
  bool unexecutable() const { return _unexecutable; }
  virtual UserCommand* clone(Customer&) const = 0;
  void set_args(const Tokens& args) { _args = args; }
  const string& get_doc_string() { return _doc_string; }

protected:
  string _doc_string;
  Tokens _args;
  bool _unexecutable;
  Customer& _customer;
};

class CommandCollection {
public:
  struct Error: public runtime_error {
    Error(const string& msg) : runtime_error(msg) {}
  };

  static UserCommand* create_command(const string name,
				     Customer& customer) {
    CommandMap::iterator i = instance()._commands.find(name);
    if(i ==  instance()._commands.end())
      throw Error("I don't know how to \"" + name + "\".");
    return i->second->clone(customer);
  }
  static void print_help() {
    const CommandMap& commands =  instance()._commands;
    for(CommandMap::const_iterator
	  i = commands.begin(); i != commands.end(); i++)
      cout << "  " << i->second->get_doc_string() << endl;
  }

private:
  static CommandCollection* _instance; 
  typedef map<string, UserCommand*> CommandMap;
  CommandMap _commands;

  CommandCollection();
  static CommandCollection& instance() {
    if(!_instance)
      _instance = new CommandCollection();
    return *_instance;
  }
};

CommandCollection* CommandCollection::_instance = 0;

// *** LOOK ***

class Look: public UserCommand {
public:
  Look(Customer& customer) : UserCommand(customer) { 
    _doc_string = "look";
    _unexecutable = false;
  }
  void execute() {
    _customer.look(true);
  }
  void unexecute() {}
  UserCommand* clone(Customer& customer) const { return new Look(customer); }
};

// *** UNDO ***

class Undo: public UserCommand {
public:
  Undo(Customer& customer) : UserCommand(customer) { 
    _doc_string = "undo";
    _unexecutable = false;
  }
  void execute() { CommandHistory::instance().undo(); }
  void unexecute() {}
  UserCommand* clone(Customer& customer) const { return new Undo(customer); }
};

// *** GO ***

class Go: public UserCommand {
public:
  Go(Customer& customer) : UserCommand(customer) { 
    _doc_string = "go [to] <mall|shop>";
  }
  void execute() { 
    int size = _args.size();
    
    int i = 0;
    if(i < size && _args.at(i) == "to")
      i++;
    if(i >= size)
      throw runtime_error("Go where?");
    string dest = _args.at(i++);
    for(; i < size; i++)
      dest += ' ' + _args.at(i);
    _last = _customer.shop;
    Shop *shop = 0;
    if(!_customer.shop) {
      Malls malls = World::instance().malls();
      for(Malls::const_iterator
	    i = malls.begin(); i != malls.end(); i++)
	if((*i)->name() == dest)
	  shop = *i;
    } else {
      const Shops& shops = _customer.shop->shops();
      for(Shops::const_iterator
	    i = shops.begin(); i != shops.end(); i++)
	if((*i)->name() == dest)
	  shop = *i;
    }
    if(!shop)
      throw runtime_error("There is no shop or mall called \""
			  + dest + "\" here.");
    _customer.go(shop);
  }
  void unexecute() {
    _customer.go(_last);
  }
  UserCommand* clone(Customer& customer) const { return new Go(customer); }

private:
  Shop* _last;
};

// *** LEAVE ***

class Leave: public UserCommand {
public:
  Leave(Customer& customer) : UserCommand(customer) { 
    _doc_string = "leave";
  }
  void execute() {
    if(!_customer.shop)
      throw runtime_error("You are not inside a shop or mall.");
    Mall *mall = dynamic_cast<Mall*>(_customer.shop);
    _sale = false;
    if(mall && !_customer.cart.products().empty()) {
      mall->check_out(_customer.cart);
      _sale = true;
    }
    _last = _customer.shop;
    _customer.leave();
  }
  void unexecute() {
    if(_sale)
      throw runtime_error("A deal is a deal. Sorry.");
    _customer.go(_last);
  }
  UserCommand* clone(Customer& customer) const { return new Leave(customer); }

private:
  Shop* _last;
  bool _sale;
};

// *** BUILD ***

class Build: public UserCommand {
public:
  Build(Customer& customer) : UserCommand(customer) {
    _doc_string = "build mall|<shop type> called <name>"; 
  }
  void execute() {
    int size = _args.size();

    // get shop type
    if(size == 0)
      throw runtime_error("Build what?");
    string type = _args.at(0);
    int i;
    for(i = 1; i < size; i++) {
      if(_args.at(i) == "called")
	break;
      type += ' ' + _args.at(i);
    }

    // get shop name
    i++;
    if(i >= size)
      throw runtime_error("You have to give the " + type + " a name.");
    string name = _args.at(i++);
    for(; i < size; i++)
      name += ' ' + _args.at(i);

    // build
    if(type == "mall") {
      if(_customer.shop)
	throw runtime_error("You can't build a mall in here!");
      _mall = new Mall(name);
      World::instance().add_mall(_mall);
    } else {
      if(!_customer.shop)
	throw runtime_error("You have to be in a mall "
			    "in order to build a shop.");
      const ShopTypes& shop_types = World::instance().shop_types();
      ShopTypes::const_iterator i = shop_types.find(type);
      if(i == shop_types.end())
	throw runtime_error("I don't know how to build a \"" + type + "\".");
      Shop *shop = i->second->clone(name, _customer.shop);
      _customer.shop->add_shop(shop);
    }
  }
  void unexecute() {
    World::instance().remove_mall(_mall);
    delete _mall;
  }
  UserCommand* clone(Customer& customer) const { return new Build(customer); }

private:
  Mall *_mall;
};

// *** HELP ***

class Help: public UserCommand {
public:
  Help(Customer& customer) : UserCommand(customer) {
    _doc_string = "help";
    _unexecutable = false;
  }
  void execute() {
    cout << endl << "Available commands:" << endl << endl;
    CommandCollection::print_help();
    const ShopTypes& shop_types = World::instance().shop_types();
    cout << endl << "Available shop types:" << endl << endl;
    for(ShopTypes::const_iterator
	  i = shop_types.begin(); i != shop_types.end(); i++)
      cout << "  " << i->first << endl;
    cout << endl
	 << "Commands can be sequenced with \"then\" "
	 << "and repeated with \"and\"." << endl
	 << "Example: \"go to body shop then take shampoo and soap\"" << endl
	 << endl;
  }
  void unexecute() {}
  UserCommand* clone(Customer& customer) const { return new Help(customer); }
};

// *** QUIT ***

class Quit: public UserCommand {
public:
  Quit(Customer& customer) : UserCommand(customer) {
    _doc_string = "quit";
    _unexecutable = false;
  }
  void execute() { exit(0); }
  void unexecute() {}
  UserCommand* clone(Customer& customer) const { return new Quit(customer); }
};

// *** LOAD ***

class Load: public UserCommand {
public:
  Load(Customer& customer) : UserCommand(customer) {
    _doc_string = "quit";
  }
  void execute() {
    int size = _args.size();
    if(size == 0)
      throw runtime_error("Open what?");
    int i = 0;
    string file = _args.at(i++);
    // This will not work well with filenames with whitespace in them,
    // but you shouldn't use those anyway :-)
    for(; i < size; i++)
      file += ' ' + _args.at(i);
    ifstream in(file.c_str());
    if(!in)
      throw runtime_error("Could not open file \"" + file + "\".");
    int line_num = 0;
    string line;      
    try {
      while(getline(in, line)) {
	line_num++;
	SyntaxTree *tree = parse(line);
	if(tree)
	  tree->interpret(_customer);
      }
    }
    catch(...) {
      cerr << "Error in " << file << "/" << line_num << ": " << line << endl;
      throw;
    }
  }
  void unexecute() {} // Todo
  UserCommand* clone(Customer& customer) const { return new Load(customer); }
};

// *** TAKE ***

class Take: public UserCommand {
public:
  Take(Customer& customer) : UserCommand(customer) {
    _doc_string = "take <product>";
  }
  void execute() {
    Shop *shop = _customer.shop;
    if(!shop)
      throw runtime_error("You are not in a shop.");
    int size = _args.size();
    if(size == 0)
      throw runtime_error("Take what?");
    int i = 0;
    string name = _args.at(i++);
    for(; i < size; i++)
      name += ' ' + _args.at(i);
    const Products& products = shop->products();
    for(Products::const_iterator 
	  i = products.begin(); i != products.end(); i++)
      if(name == (*i)->name) {
	shop->put_in_cart(_customer.cart, *i);
	_product = *i;
	return;
      }
    throw runtime_error("There is no \"" + name + "\"");
  }
  void unexecute() {
    _customer.shop->remove_from_cart(_customer.cart, _product);
  }
  UserCommand* clone(Customer& customer) const { return new Take(customer); }
private:
  const Product* _product;
};

CommandCollection::CommandCollection() {
  static Customer dummy("John Doe");
  _commands["look"] = new Look(dummy);
  _commands["help"] = new Help(dummy);
  _commands["quit"] = new Quit(dummy);
  _commands["build"] = new Build(dummy);  
  _commands["undo"] = new Undo(dummy);
  _commands["go"] = new Go(dummy);
  _commands["leave"] = new Leave(dummy);
  _commands["load"] = new Load(dummy);
  _commands["take"] = new Take(dummy);
}

void Sequence::interpret(Customer& customer) {
  int j = 1;
  for(Trees::const_iterator
	i = _children.begin(); i != _children.end(); i++, j++) {
    //    cout << j << ": " << endl;
    (*i)->interpret(customer);
  }
}

void CommandRep::interpret(Customer& customer) {
  if(_children.empty())
    _children.push_back(new Arguments(Tokens()));
  for(Trees::const_iterator
 	i = _children.begin(); i != _children.end(); i++) {
    UserCommand *cmd = CommandCollection::create_command(_name, customer);
    cmd->set_args(dynamic_cast<Arguments*>(*i)->tokens());
    cmd->execute();
    if(cmd->unexecutable())
      CommandHistory::instance().push_command(cmd, false);
    else
      delete cmd;
  }
//   if(_name == "look")
//     customer.look(true);
//   cout << "  " << _name << endl;
//   for(Trees::const_iterator
// 	i = _children.begin(); i != _children.end(); i++) {
//     cout << "    ";
//     (*i)->interpret(customer);
//   }
}

void Arguments::interpret(Customer& customer) {
  for(Tokens::const_iterator
	i = _tokens.begin(); i != _tokens.end(); i++) {
    cout << *i << " ";
  }
  cout << endl;
}
