#include <sys/select.h>
#include <stdio.h> // perror
#include <fcntl.h> // open
#include <stdlib.h> // malloc, rand
#include <glib.h>
#include <unistd.h> // chdir, stat
#include <sys/types.h> // time_t
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <libgnomecanvas/libgnomecanvas.h>
#include <math.h> // sqrt, fabs, exp
#include <time.h>
#include <string.h> // strlen, strstr, strcmp

#define FOREACH(x, array) for(int x = 0; x < array->len; x++)

// ****** VECTOR ARITHMETIC ******

struct Vector { double x, y; }; typedef struct Vector Vector;
Vector add(Vector a, Vector b) { return (Vector) {a.x + b.x, a.y + b.y}; }
Vector sub(Vector a, Vector b) { return (Vector) {a.x - b.x, a.y - b.y}; }
Vector mul(double f, Vector v) { return (Vector) {f * v.x, f * v.y}; }
double len(Vector v) { return sqrt(v.x * v.x + v.y * v.y); }

// ****** RULE NODE DEFINITION ******

struct Rule {
  char *target, *cmd;
  GPtrArray *deps;
  enum { PENDING, RUNNING, OK, FAILED } status;
  GString *msg;
  
  GnomeCanvasGroup *group;
  GnomeCanvasText *text;
  GnomeCanvasRE *indicator;
  GPtrArray *arrows;
  Vector pos, new_pos;
};
typedef struct Rule Rule;

/// allocate and initalize a Rule object
Rule* make_rule(char *target, char *cmd) {
  Rule *rule = (Rule*) malloc(sizeof(Rule));
  rule->target = target;
  rule->cmd = cmd;
  rule->deps = g_ptr_array_new();
  rule->status = PENDING;
  rule->msg = g_string_new(0);
  return rule;
}

/// add a dependency
void add_dep(Rule *rule, char *dep) { g_ptr_array_add(rule->deps, dep); }

/// returns the position of the GnomeCanvas Group of a Rule
Vector get_pos(Rule *rule) {
  Vector v;
  g_object_get(rule->group, "x", &v.x, "y", &v.y, NULL);
  return v;
}

/// if true, the target is expected to exist a priori
int pre_existing(Rule *rule) { return rule->deps->len == 0 && rule->cmd == 0; }

// ****** MAKEFILE PARSER ******

/// remove trailing newline
GString* nonl(GString* s) {
  if(s->len && s->str[s->len - 1] == '\n')
    s->str[s->len - 1] = 0;
  return s;
}

/// show error message and quit
void parse_error(char *msg) {
  fprintf(stderr, "Parse error: %s\n", msg);
  exit(2);
}

/// substitute $(MAKEFILE_STYLE) variables in string
GString* subst_vars(GHashTable *vars, char *s) {
  GString *result = g_string_new(s);
  char *start = result->str;
  while((start = strstr(start, "$("))) {
    char *end = strstr(start, ")");
    if(!end)
      parse_error("Variable missing closing parenthesis");
    int start_pos = start - result->str;
    char *name_start = start + 2;
    int name_len = end - name_start;
    GString *name = g_string_new_len(name_start, name_len);
    char *value = g_hash_table_lookup(vars, name->str);
    g_string_erase(result, start_pos, name_len + 3);
    if(value)
      g_string_insert(result, start_pos, value);
    start = result->str + start_pos + (value ? strlen(value) : 0); 
  }
  return result;
}

/// remove everything after '#'
void uncomment(GString *s) {
  gchar **tokens = g_strsplit(s->str, "#", 2);
  g_string_set_size(s, 0);
  if(tokens && tokens[0])
    g_string_append(s, tokens[0]);
  g_strfreev(tokens);
}

/// parse a Makefile rule from after the colon
Rule* parse_rule(GString *line, GIOChannel *in, GHashTable *vars) {
  gchar **tokens = g_strsplit(line->str, ":", 2);
  GString *depstring = subst_vars(vars, tokens[1]);
  GString *target = subst_vars(vars, tokens[0]);
  g_strfreev(tokens);
  tokens = g_strsplit(depstring->str, " ", 0);
  g_string_free(depstring, TRUE);
  GString *cmd = g_string_new(0);
  int first = TRUE;
  while(g_io_channel_read_line_string(in, line, 0, 0) == G_IO_STATUS_NORMAL
        && line->str[0] == '\t') {
    uncomment(nonl(line));
    g_string_erase(line, 0, 1); // remove tab
    if(!first)
      g_string_append(cmd, " && ");
    first = FALSE;
    g_string_append(cmd, line->str);
  }
  Rule *rule = 0;
  // ignore special targets
  if(!(strcmp(target->str, "clean") == 0
       || (strcmp(target->str, "run") == 0))) {
    rule = make_rule(target->str, subst_vars(vars, cmd->str)->str);
    for(gchar **token = tokens; token && *token; token++)
      if(**token)
	add_dep(rule, g_string_new(*token)->str);
  }
  g_strfreev(tokens);
  return rule;
}

/// remove initial and trailing space chars from string
GString* nows(GString *s) {
  while(s->str[0] == ' ')
    g_string_erase(s, 0, 1);
  while(s->len > 0 && s->str[s->len - 1] == ' ')
    g_string_erase(s, s->len - 1, 1);
  return s;
}

/// parse a variable definition from after the '='
void parse_var_def(GHashTable *vars, GString *line) {
  gchar **tokens = g_strsplit(line->str, "=", 2);
  GString *varname = nows(subst_vars(vars, tokens[0]));
  GString *varvalue = nows(subst_vars(vars, tokens[1]));
  g_hash_table_insert(vars, varname->str, varvalue->str);
  g_strfreev(tokens);
}

/// parse a makefile and return a list of rules connected to each other
GPtrArray* parse_rules(int makefile) {
  GPtrArray *rules = g_ptr_array_new();
  GIOChannel *in = g_io_channel_unix_new(makefile);
  GString *line = g_string_new(0);
  GHashTable *vars = g_hash_table_new(g_str_hash, g_str_equal);
  while(g_io_channel_read_line_string(in, line, 0, 0)== G_IO_STATUS_NORMAL) {
  loop_start:
    uncomment(nonl(line));
    char *colonpos = strstr(line->str, ":");
    char *equalpos = strstr(line->str, "=");
    if(!equalpos && !colonpos)
      continue; // todo: don't ignore lines with neither defs or rules
    if(equalpos && (!colonpos || colonpos > equalpos))
      parse_var_def(vars, line);
    else {
      Rule *rule = parse_rule(line, in, vars);
      if(rule)
	g_ptr_array_add(rules, rule);
      goto loop_start; // yuck!
    }
  }
  g_hash_table_destroy(vars);

  // substitute target names for rules in Rule.deps
  GHashTable *rules_table = g_hash_table_new(g_str_hash, g_str_equal);
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    g_hash_table_insert(rules_table, rule->target, rule);
  }
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    FOREACH(i, rule->deps) {
      char *dep = g_ptr_array_index(rule->deps, i);
      Rule *deprule = g_hash_table_lookup(rules_table, dep);
      if(deprule) {
        rule->deps->pdata[i] = deprule;
        free(dep);
      } else {
        rule->deps->pdata[i] = make_rule(dep, 0);
        g_ptr_array_add(rules, rule->deps->pdata[i]);
        g_hash_table_insert(rules_table, dep, rule->deps->pdata[i]);
      }
    }
  }
  g_hash_table_destroy(rules_table);
  return rules;
}

// ****** BUILD AND GUI CODE ******

/// get the modification time of a file
time_t mtime(const char *filename) {
  struct stat buf;
  if(stat(filename, &buf) != 0)
    return 0;
  return buf.st_mtime;
}

#define UPDATE_MAX(max, val) max = val > max ? val : max; 

/// sleep for ms milliseconds
void sleepms(long ms) {
  select(0, 0, 0, 0, & (struct timeval) { .tv_usec = 1000 * ms });
}

static const int delay = 100; /// thread loop delay in ms

/// set Rule status and update display
void set_status(Rule *rule, int status) {
  rule->status = status;
  char *color, *msg;
  static char
    pending_c[] = "yellow", running_c[] = "white", failed_c[] = "red",
    ok_c[] = "green", pending_m[] = "\nPENDING", running_m[] = "\nRUNNING",
    failed_m[] = "\nFAILED", ok_m[] = "\nOK";
  switch(status) {
  case PENDING: color = pending_c; msg = pending_m; break;
  case RUNNING: color = running_c; msg = running_m; break;
  case FAILED: color = failed_c; msg = failed_m; break;
  case OK: color = ok_c; msg = ok_m; break;
  }
  gdk_threads_enter();
  g_string_set_size(rule->msg, 0);
  g_string_append(rule->msg, rule->target);
  g_string_append(rule->msg, msg);
  gnome_canvas_item_set(GNOME_CANVAS_ITEM(rule->indicator),
                        "fill-color", color, NULL);
  gnome_canvas_item_set(GNOME_CANVAS_ITEM(rule->text),
                        "text", rule->msg->str, NULL);
  gdk_threads_leave();
}

/// this function runs in a thread for each node in the Rule graph
/// and performs the commands in the rule
gpointer run(gpointer r) {
  Rule *rule = (Rule*) r;
  int done = 0;
  time_t latest_dep = 0;
  set_status(rule, PENDING);
  while(!done) {
    sleepms(delay);
    done = 1;
    FOREACH(i, rule->deps) {
      Rule *deprule = (Rule*) g_ptr_array_index(rule->deps, i);
      
      if(deprule->status == FAILED) {
        set_status(rule, FAILED);
        return 0;
      } else if(deprule->status != OK) {
        done = 0;
      } else {
        UPDATE_MAX(latest_dep, mtime(deprule->target));
      }
    }
  }
  // no need to rebuild if the target is newer that the deps
  if(mtime(rule->target) > latest_dep) {
    set_status(rule, OK);
    return 0;
  }
  if(!rule->cmd) { // no rule - assume pre-existing
    // should have been caught by the mtime call above
    fprintf(stderr, "File %s missing", rule->target);
    set_status(rule, FAILED);
    return 0;
  }
  set_status(rule, RUNNING);
  fprintf(stderr, "Running: %s\n", rule->cmd);
  if(system(rule->cmd) == 0) {
    set_status(rule, OK);
    fprintf(stderr, "Done: %s\n", rule->cmd);
  } else {
    set_status(rule, FAILED);
    fprintf(stderr, "Failed: %s\n", rule->cmd);
  }
  return 0;
}

/// build all targets (in separate threads)
int build_all(GPtrArray *targets) {
  GPtrArray *threads = g_ptr_array_new();
  FOREACH(i, targets)
    g_ptr_array_add(threads, g_thread_create(run, g_ptr_array_index(targets, i),
                                             FALSE, 0));
  return 1;
}

/// change position of a node in the gui
void move_rule_object(Rule *rule, Vector pos) {
  Vector delta = sub(pos, get_pos(rule));
  gnome_canvas_item_move(GNOME_CANVAS_ITEM(rule->group), delta.x, delta.y);
}

/// node radius
const double n_radius = 15.0;

/// make sure the arrows hit their marks
void update_arrows(Rule *rule) {
  Vector rulepos = get_pos(rule);
  FOREACH(i, rule->arrows) {
    Rule *deprule = (Rule*) g_ptr_array_index(rule->deps, i);
    GnomeCanvasLine *arrow =
      GNOME_CANVAS_LINE(g_ptr_array_index(rule->arrows, i));
    Vector deppos = get_pos(deprule);
    Vector diff = sub(rulepos, deppos);
    double l = len(diff);
    diff = add(deppos, mul((l - n_radius)/l, diff));
    GnomeCanvasPoints *points;
    g_object_get(arrow, "points", &points, NULL);
    points->coords[0] = diff.x; points->coords[1] = diff.y;
    points->coords[2] = deppos.x; points->coords[3] = deppos.y;
    gnome_canvas_item_set(GNOME_CANVAS_ITEM(arrow), "points", points, NULL);
  }
}

/// create the GnomeCanvas items of a node
void make_rule_object(GnomeCanvas *canvas, Rule *rule, double x, double y) {
  rule->pos = (Vector) {x, y};
  rule->group = GNOME_CANVAS_GROUP
    (gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
                           gnome_canvas_group_get_type(),
                           "x", x, "y", y, NULL));
  rule->text = GNOME_CANVAS_TEXT
    (gnome_canvas_item_new(rule->group, gnome_canvas_text_get_type(),
                           "fill-color", "blue",
                           "x", 0.0, "y", 20.0, "anchor", GTK_ANCHOR_NORTH,
                           "text", rule->target, "size", 8000, NULL));
  rule->indicator = GNOME_CANVAS_RE
    (gnome_canvas_item_new(rule->group,
                           !pre_existing(rule)
                           ? gnome_canvas_ellipse_get_type()
                           : gnome_canvas_rect_get_type(),
                           "fill-color", "gray",
                           "width-pixels", 3, "outline-color", "black",
                           "x1", -n_radius, "x2", n_radius,
                           "y1", -n_radius, "y2", n_radius,
                           NULL));
  rule->arrows = g_ptr_array_new();
  FOREACH(i, rule->deps) {
    GnomeCanvasPoints *points = gnome_canvas_points_new(2);
    // the points have to be initialized to something
    points->coords[0] = 10; points->coords[1] = 10;
    points->coords[2] = 20; points->coords[3] = 30;
    GnomeCanvasItem *item = 
      gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
                            gnome_canvas_line_get_type(),
                            "width-pixels", 2, "fill-color", "black",
                            "points", points, "first-arrowhead", TRUE,
                            "arrow-shape-a", 10.0, "arrow-shape-b", 15.0,
                            "arrow-shape-c", 5.0,
                            NULL);
    g_ptr_array_add(rule->arrows, item);
    gnome_canvas_item_lower_to_bottom(item);
  }
}

float _rand() { return ((float) rand() / RAND_MAX - 0.5); }


/// create canvas items for all rules (in random locations)
void make_rule_objects(GnomeCanvas *canvas, GPtrArray *rules) {
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    float w = 100;
    make_rule_object(canvas, rule, w * _rand(), w * _rand());
  }
}

int radius = 400;

/// Tthe graph is self-organizing, by minimizing its own "energy". A
/// minimizing step is performed by this ad-hoc function.
void update_graph(GPtrArray *rules) {
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    rule->new_pos = rule->pos;
  }
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    // don't leave the window
    rule->new_pos = sub(rule->new_pos, mul(0.03, rule->pos));
    // repulsive force
    FOREACH(j, rules) {
      if(j == i) continue;
      Rule *orule = (Rule*) g_ptr_array_index(rules, j);
      Vector diff = sub(rule->pos, orule->pos);
      double l = len(diff);
      Vector f = mul(100.0/(0.1 + l * l), diff);
      rule->new_pos = add(rule->new_pos, f);
    }
    // spring force in deps
    FOREACH(j, rule->deps) {
      Rule *dep = (Rule*) g_ptr_array_index(rule->deps, j);
      Vector diff = sub(dep->pos, rule->pos);
      Vector f = mul(0.005, diff);
      rule->new_pos = add(rule->new_pos, f);
      dep->new_pos = sub(dep->new_pos, f);
    }
  }
}

/// actually move the objects to their new positions
void apply_update(GPtrArray *rules, int redraw) {
  FOREACH(i, rules) {
    Rule *rule = (Rule*) g_ptr_array_index(rules, i);
    rule->pos = rule->new_pos;
    if(redraw)
      move_rule_object(rule, rule->pos);
  }
  FOREACH(i, rules)
    if(redraw)
      update_arrows((Rule*) g_ptr_array_index(rules, i));
}

/// the graph organizer runs in a separate thread
gpointer graph_updater(gpointer r) {
  GPtrArray *rules = (GPtrArray*) r;
  for(int i = 0; i < 20; i++) {
    for(int j = 0; j < 10; j++) {
      update_graph(rules);
      apply_update(rules, FALSE);
    }
    gdk_threads_enter();
    apply_update(rules, TRUE);
    gdk_threads_leave();
    sleepms(50);
  }
  return 0;
}

/// quit using the wm
void destroy(GtkWidget *w, gpointer data) { gtk_main_quit (); }
/// push button to start
void build(GtkWidget *w, gpointer rules) { build_all((GPtrArray*) rules); }

/// build gui
void create_window(GPtrArray *rules) {
  GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  GtkWidget *button = gtk_button_new_with_label("Bygg");
  GtkWidget *box = gtk_vbox_new(FALSE, 0);
  // The antialiased version doesn't work very well at Stacken
  GtkWidget *canvas = FALSE ? gnome_canvas_new_aa() : gnome_canvas_new();
  GtkWidget *scroller = gtk_scrolled_window_new
    (gtk_layout_get_hadjustment(GTK_LAYOUT(canvas)),
     gtk_layout_get_vadjustment(GTK_LAYOUT(canvas)));
  gtk_widget_set_size_request(canvas, 2 * radius, 2 * radius);
  int sr = 2 * radius;
  gnome_canvas_set_scroll_region(GNOME_CANVAS(canvas), -sr, -sr, sr, sr);
  gnome_canvas_scroll_to(GNOME_CANVAS(canvas), sr/2, sr/2);
  gtk_container_add(GTK_CONTAINER(window), box);
  gtk_box_pack_start(GTK_BOX(box), scroller, TRUE, TRUE, 0);
  gtk_container_add(GTK_CONTAINER(scroller), canvas);
  gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
  gtk_widget_show_all(window);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(build),
                   rules);
  g_signal_connect(G_OBJECT(window), "destroy",
                   G_CALLBACK(destroy), NULL);

  make_rule_objects(GNOME_CANVAS(canvas), rules);
}

int main(int argc, char **argv) {
  if(argc == 2 && chdir(argv[1]) != 0) { perror(0); exit(1); }
  int makefile = open("Makefile", O_RDONLY);
  if(makefile == -1) { perror("Could not open Makefile. "); exit(1); }
  GPtrArray *rules = parse_rules(makefile);
  g_thread_init(0); gdk_threads_init(); gtk_init(&argc, &argv);
  create_window(rules);
  g_thread_create(graph_updater, rules, FALSE, 0); // graph organizer thread
  gdk_threads_enter();
  gtk_main(); // main event loop
  gdk_threads_leave();
}
